# Name: xdp/data.py
# Purpose: XDP vector-oriented data manipulation
# Author: Ken McIvor <mcivken@iit.edu>
#
# Copyright 2003-2004 Illinois Institute of Technology
#
# See the file "LICENSE" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.

from __future__ import nested_scopes


"""
Data manipulation primitives that operate on Numerical
Python arrays.
"""


import math
import bisect
import Numeric


def grid(start, stop, stepSize):
    """
    Create a vector which can be used as an abscissa for interpolation.

    @param start: first value in the grid
    @type  start: C{int} or C{float}
    @param stop: last value in the grid
    @type  stop: C{int} or C{float}
    @param stepSize: difference between values in the grid
    @type  stepSize: C{int}, C{float}, or C{complex}
    @return: vector of values going from C{start} to C{stop} by {stepSize}
    @rtype:  C{Numeric} vector
    """

    r = Numeric.arange(start, stop + stepSize, stepSize)
    while r[-1] > stop:
        r = r[:-1]
    if r[-1] < stop:
        r = Numeric.resize(r, (r.shape[0]+1, ) )
        r[-1] = stop
    return r


def minmax(*vectors):
    """
    Find the greatest minimum and least maximum of a series of vectors.

    Example::

        >>> x = [1,2,3,4,5]
        >>> y = [0,2,4,6,8]
        >>> print xdp.range.minmax(x, y)
        (1, 5)

    @param vectors: series of vectors
    @type  vectors: C{n-tuple} of C{Numeric} vectors
    @return: the greatest minimum and least maximum of C{vectors}
    @rtype:  C{2-tuple} of the form C{(greatest_minimum, least_maximum)}
    """
    return greatestMin(*vectors), leastMax(*vectors)


def greatestMin(*vectors):
    """
    Find the greatest minimum value of a series of vectors.

    @param vectors: series of vectors
    @type  vectors: C{n-tuple} of C{Numeric} vectors
    @return: greatest minimum value of C{vectors} or zero if C{vectors} is
             empty
    @rtype:  C{int}, C{float}, or C{complex}
    """

    if not vectors:
        return 0.0
    else:
        return max([ min(v) for v in vectors ])


def leastMax(*vectors):
    """
    Find the least maximum value of a series of vectors.

    @param vectors: series of vectors
    @type  vectors: C{n-tuple} of C{Numeric} vectors
    @return: least maximum value of C{vectors} or zero if C{vectors} is empty
    @rtype:  C{int}, C{float}, or C{complex}
    """

    if not vectors:
        return 0.0
    else:
        return min([ max(v) for v in vectors ])


def checkRange(Min, Max, start, end):
    """
    Perform range checking and assist in implementing autoranging.

    The values of C{start} and C{end} are checked for validity against the
    range minimum C{Min} and the range maximum C{Max}.  If C{start} or C{end}
    is C{None} then the value of C{Min} or C{Max}, repectively, will be
    returned in its place.

    @param Min: minimum valid value in the range
    @type  Min: C{int} or C{float}
    @param Max: maximum valid value in the range
    @type  Max: C{int} or C{float}
    @param start: start of the range
    @type  start: C{int}, C{float}, or C{None}
    @param end: end of the range
    @type  end: C{int}, C{float}, or C{None}
    @return: final start and end values
    @rtype:  C{2-tuple} of the form C{(start, end)}

    @raise ValueError: either C{start} or C{end} is outside of the valid range
    """

    if start is not None:
        if start > Max:
            raise ValueError('range start %s is too large (maximum is %s)'
                % (start, Max))
        elif start < Min:
            raise ValueError('range start %s is too small (minimum is %s)'
                % (start, Min))
        else:
            Min = start

    if end is not None:
        if end > Max:
            raise ValueError('range end %s is too large (maximum is %s)'
                % (end, Max))
        elif end < Min:
            raise ValueError('range end %s is too small (minimum is %s)'
                % (end, Min))
        else:
            Max = end

    return Min, Max


def trim(x, y, start, end):
    """
    Trim vectors to fit a new range of independent values.

    All of the M{x}/M{y} pairs outside of the range of C{start} to C{end} are
    discarded.  Either C{start} or C{end} may be C{None}, which indicates that
    that portion of the range should not be trimmed.

    @param x: independent values
    @type  x: C{Numeric} vector
    @param y: dependent values
    @type  y: C{Numeric} vector or array
    @param start: beginning of the new range
    @type  start: C{int}, C{float}, or C{None}
    @param end: end of the new range
    @type  end: C{int}, C{float}, or C{None}
    @return: trimmed vectors
    @rtype:  C{2-tuple} of the form C{(trimmed_x, trimmed_y)}

    @raise ValueError: either C{start} or C{end} is outside the valid range
                       for C{x}

    @note: Due to the implementation of L{Numeric}, the trimmed vectors
    reference the same data as the original vectors.  Subsequent changes made
    to the trimmed vectors will be reflected in the values of the original
    vectors.  If this is undesirable, use the C{copy} method of the vectors to
    create an independent copy of the data.
    """
    Min = min(x)
    Max = max(x)
    checkRange(Min, Max, start, end)

    if start is not None and start > Min:
        xs = bisect.bisect_left(x, start)
        x = x[xs:]
        if y is not None:
            y = y[xs:]

    if end is not None and end < Max:
        xe = bisect.bisect_right(x, end)
        x = x[0:xe+1]
        if y is not None:
            y = y[0:xe+1]

    return (x, y)


def clean(v, *vectors):
    """
    Remove recurring values from one or more vectors.

    Recurring values are removed from the vector C{v} and the corresponding
    elements are removed from elements of C{vectors}

    @param v: vector to check for recurring values
    @type  v: C{Numeric} vector
    @param vectors: additional vectors to clean
    @type  vectors: C{n-tuple} of C{Numeric} vectors
    @return: vectors with recurring values in C{v} removed
    @rtype : C{n-tuple} of C{Numeric} vectors

    @raise ValueError: not all elements of C{vectors} are vectors with the same
                       length as C{v}
    """
    v = Numeric.asarray(v)
    vectors = [ Numeric.asarray(x) for x in vectors ]
    shapes = [v.shape] + [ x.shape for x in vectors ]

    if not Numeric.alltrue(Numeric.equal([len(x) for x in shapes], 1)):
        raise ValueError('All arguments must be vectors')
    elif not Numeric.alltrue(Numeric.equal(shapes, shapes[0])):
        raise ValueError('All arguments must be vectors of the same length')

    indices = {}
    for i in range(0, v.shape[0]):
        indices[v[i]] = i

    elements = indices.keys()
    elements.sort()
    indices = [ indices[x] for x in elements ]
    results = []
    for vector in [v] + vectors:
        cleaned = Numeric.zeros((len(indices),), vector.typecode())
        for i in range(0, len(indices)):
            cleaned[i] = vector[indices[i]]
        results.append(cleaned)
    return results


def coljoin(*vectors):
    """
    Join two or more vectors together as the columns of a matrix.

    @param vectors: vectors to join
    @type  vectors: C{n-tuple} of C{Numeric} vectors
    @return: matrix with C{vectors} as the columns
    @rtype:  C{Numeric} matrix or C{None}

    @raise ValueError: not all elements of C{vectors} are vectors with the same
                       length

    @note: The function returns None when C{vectors == ()}.
    """
    if not vectors:
        return None

    vectors = [ Numeric.asarray(x) for x in vectors ]
    columns = len(vectors)
    shapes = [ x.shape for x in vectors ]

    if not Numeric.alltrue(Numeric.equal([len(x) for x in shapes], 1)):
        raise ValueError('All arguments must be vectors')
    elif not Numeric.alltrue(Numeric.equal(shapes, shapes[0])):
        raise ValueError('All arguments must be vectors of the same length')

    matrix = Numeric.zeros(shapes[0]+(columns,), 'd')
    for i in range(0, columns):
        matrix[:,i] = vectors[i]
    return matrix


class RebinError(Exception):
    """
    Exception indicating that an error occured during rebinning.
    """
    pass


def rebin(data, eColumn, Eo, offset, estep, kstep):
    """
    Reduce XAFS data by binning it's pre-edge and XAFS regions.
   
    All points for which the energy is greater than M{C{Eo}+C{offset}} are
    reduced to a new set of points in which the energy varies regularly with
    respect to k-space, which increases by C{k-step} per point.

    If {e-step} is not C{None}, then all points for which the energy is less
    than M{C{Eo}+C{offset}} are reduced to a new set of points in which the
    energy varies regularly with an increas of C{e-step} per point.

    @param data: XAFS data
    @type  data: C{Numeric} matrix
    @param eColumn: index of the I{energy} column in C{data}
    @type  eColumn: C{int}
    @param Eo: edge energy
    @type  Eo: C{int} or C{float}
    @param offset: rebin this much before and after Eo
    @type  offset: C{int} or C{float}
    @param estep: bin size of pre-edge region
    @type  estep: C{int} or C{float} or C{None} to disable pre-edge rebinning
    @param kstep: bin size of XAFS region in k-space
    @type  kstep: C{int} or C{float}
    @return: matrix of rebinned data
    @rtype:  C{Numeric} matrix

    @raise RebinError: C{Eo} or C{offset} is too large
    """

    # acquire the energy vector of the data and find its minimum and maximum
    E = data[:,eColumn]
    Min, Max = min(E), max(E)

    if Max <= Eo:
        raise RebinError('there are no energy values greater than Eo')

    #
    # Setup to rebin the pre-edge region
    #

    # start rebinning points in the pre-edge region at Eo-offset
    preedge_end = Eo - offset

    # should the pre-edge region be rebinned? does it have any data points?
    if estep is None or preedge_end < Min:
        preedge_endPoint = 0
        preedge_pointCount = 0
        preedge_eRange = None
        preedge_dataRange = None
        egrid = None
        rebinned_preedge = None
    else:
        # find the index of the energy vector which corresponds to the end of
        # the pre-edge region
        preedge_endPoint = bisect.bisect_right(E, preedge_end)

        # slice the energy vector and data matrix to get the points for
        # rebinning
        preedge_eRange = E[0:preedge_endPoint]
        preedge_dataRange = data[0:preedge_endPoint, :]

        # Generate the regular abscissa that the pre-edge region will be
        # rebined to.
        #
        # grid(start, end, step) returns a vector of values from 'start' to
        # 'end' by 'step', inclusive.
        egrid = grid(Min, preedge_end, estep)

        # drop the last bin if it's too narrow, to avoid pathological cases
        if egrid[-1] - egrid[-2] < estep/2.0:
            preedge_end = egrid[-2]
            preedge_endPoint = bisect.bisect_right(E, preedge_end)
            preedge_eRange = E[0:preedge_endPoint]
            preedge_dataRange = data[0:preedge_endPoint, :]
            egrid = egrid[0:-1]

        # rebin the pre-edge region
        rebinned_preedge = rebin_data(preedge_eRange, preedge_dataRange, egrid)
        preedge_pointCount = rebinned_preedge.shape[0]

    #
    # Setup to rebin the XAFS region
    #

    # start rebinning points in the XAFS region at Eo+offset
    xafs_start = Eo + offset

    if Max <= xafs_start:
        raise RebinError(
            'the XAFS region contains no points (`offset\' is too large)')

    # find the index of the energy vector which corresponds to the start point
    xafs_startPoint = bisect.bisect_left(E, xafs_start)

    # slice the energy vector and data matrix to get the points for rebinning,
    # converting from energy to k-space in the process
    xafs_kRange = 0.512*Numeric.sqrt(E[xafs_startPoint:] - Eo)
    xafs_dataRange = data[xafs_startPoint:, :]

    # Generate the regular k-space abscissa that will be rebined to.
    kstart = 0.512*math.sqrt(xafs_start - Eo)
    kend = 0.512*math.sqrt(Max - Eo)
    kgrid = grid(kstart, kend, kstep)

    # rebin the XAFS region
    rebinned_xafs = rebin_data(xafs_kRange, xafs_dataRange, kgrid)
    xafs_pointCount = rebinned_xafs.shape[0]

    #
    # Assemble the resulting data matrix
    #

    # Determine the number of points in the edge region.
    edge_startPoint = preedge_endPoint
    edge_endPoint = xafs_startPoint
    edge_pointCount = data[edge_startPoint:edge_endPoint].shape[0]

    # create a new matrix to hold the final result
    rowCount = preedge_pointCount + edge_pointCount + xafs_pointCount
    columnCount = data.shape[1]
    result = Numeric.zeros((rowCount, columnCount), data.typecode())    

    # if applicable, store the rebinned pre-edge region
    if rebinned_preedge is not None:
        s = 0
        e = preedge_pointCount 
        result[s:e, :] = rebinned_preedge

    # store the unmodified edge region
    s = preedge_pointCount
    e = s + edge_pointCount
    result[s:e, :] = data[edge_startPoint:edge_endPoint, :]

    # store the rebinned XAFS region
    s = preedge_pointCount + edge_pointCount
    e = s + xafs_pointCount
    result[s:e, :] = rebinned_xafs

    return result


def rebin_data(x, data, grid):
    """
    The data reduction algorithm used by rebin().

    @param x: values of the independent variable from C{data}
    @type  x: C{Numeric} vector
    @param data: data to be reduced
    @type  data: C{Numeric} matrix
    @param grid: regular abscissa of bin values
    @type grid: C{Numeric} vector
    @rtype:  C{Numeric} matrix

    @raise RebinError: one of the bins contained no data
    """

    columnCount = data.shape[1]
    binBoundaryCount = grid.shape[0]
    binnedValueCount = grid.shape[0] - 1

    # newData is a matrix which is binnedValueCount rows long and contains the
    # same number of columns as the data matrix.  It accumulates the sums of
    # the points which fall in each bin
    newData = Numeric.zeros((binnedValueCount, columnCount), data.typecode())

    # binCounts is a vector which accumulates the number of points that fall
    # into each bin
    binCounts = Numeric.zeros((binnedValueCount,), data.typecode())

    for i in range(0, x.shape[0]):
        # locate the index of the bin in which the independent value falls
        where = bisect.bisect_right(grid, x[i]) - 1
        if where == newData.shape[0]:
            where = where - 1

        # bin the data
        newData[where] += data[i]
        binCounts[where] += 1

    # average the binned points
    for i in range(0, binnedValueCount):
        if binCounts[i] != 0.0:
            newData[i,:] /= binCounts[i]

    return newData


def old_rebin(data, eColumn, Eo, after, kstep):
    """Reduce XAFS data by binning it in k-space.

    All points for which the energy is greater than M{C{Eo}+C{after}} are
    reduced to a new set of points in which the energy varies regularly with
    respect to k-space, which increases by C{k-step} per point.

    @param data: XAFS data
    @type  data: C{Numeric} matrix
    @param eColumn: index of the I{energy} column
    @type  eColumn: C{int}
    @param Eo: edge energy
    @type  Eo: C{int} or C{float}
    @param after: rebin this much after Eo
    @type  after: C{int} or C{float}
    @param kstep: bin size in k-space
    @type  kstep: C{int} or C{float}
    @return: matrix of rebinned data
    @rtype:  C{Numeric} matrix

    @raise RebinError: C{Eo} or C{after} is too large, or on of the k-space
                       bins contained no data
    """

    # acquire the energy vector of the data and find its minimum and maximum
    E = data[:,eColumn]
    Min, Max = min(E), max(E)

    if Max <= Eo:
        raise RebinError('no energy values greater than Eo')

    # start rebinning points at Eo+after
    start = Eo + after

    if Max <= start:
        raise RebinError('the value of \'after\' is too large')

    # find the index of the energy vector which corresponds to the start point
    startPoint = bisect.bisect_left(E, start)

    # slice the energy vector and data matrix to get the points for rebinning
    eRange = E[startPoint:]
    dataRange = data[startPoint:, :]

    def k(E, Eo=Eo):
        """Calculate K for some energy level E, using the edge energy argument
        passed to rebin().
        """
        return 0.512 * math.sqrt(E - Eo)

    # Generate the regular k-space abscissa that will be rebined to.
    # grid(start, end, step) returns a vector of values from 'start' to 'end'
    # by 'step', inclusive.
    kgrid = grid(k(start), k(Max), kstep)

    # newData is matrix which is BINCOUNT-1 rows long and contains the same
    # number of columns as the data matrix.  It accumulates the sums of the
    # points which fall in each bin
    newData = Numeric.zeros((kgrid.shape[0]-1, data.shape[1]), data.typecode())

    # binCounts is a vector of BINCOUNT elements, which accumulates the number
    # of points that fall in each bin
    binCounts = Numeric.zeros( (newData.shape[0],) , data.typecode())

    for i in range(0, eRange.shape[0]):
        # locate the index of the bin in which the current energy level falls
        where = bisect.bisect_right(kgrid, k(eRange[i])) - 1
        if where == newData.shape[0]:
            where = where - 1

        # bin the data
        newData[where] += dataRange[i]
        binCounts[where] += 1

    # check that all of the bins contain data
    for i in range(0, newData.shape[0]):
        if binCounts[i] == 0:
            raise RebinError, ('bin %d (%.4f < k < %.4f) is empty'
                % (i, kgrid[i], kgrid[i+1]))

    # average the binned points
    for i in range(0, newData.shape[0]):
        newData[i,:] /= binCounts[i]

    # The length of E[0:startPoint] is the number of points which were not
    # rebinned.  The result matrix will contain that many rows plus one row for
    # each bin.
    Len = E[0:startPoint].shape[0] + newData.shape[0]

    # create a new matrix to hold the final result
    result = Numeric.zeros((Len, data.shape[1]), data.typecode())   

    # copy unrebinned and rebinned data into the result matrix
    result[0:startPoint, :] = data[0:startPoint, :]
    result[startPoint:, :] = newData

    return result

