# Name: xdp/dataset.py
# Purpose: High-level dataset representation and manipulation
# Author: Ken McIvor <mcivor@iit.edu>
#
# Copyright 2003-2009 Illinois Institute of Technology
#
# See the file "LICENSE" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.

"""
High-level dataset representation and manipulation.
"""


import re
import sys
import math
import types
import Numeric

from xdp.errors import *


class Data:
    """
    Store columns of data and their associated names.

    These columns can be maniplulated by the get(), add(), and set()
    methods or by accessing the columns as though they are attributes of the
    object.

    @cvar COLNAMES: used by C{Data.evaluate} to match and replace the I{@X}
                    column names in expressions
    @type COLNAMES: compiled regular expression
    @cvar COLNUMS: used by C{Data.evaluate} to match and replace the I{@X}
                   column numbers in expressions
    @type COLNUMS: compiled regular expression

    @cvar EVAL_NAMESPACE: used by C{Data.evaluate} as the default namespace for
                          evaluating Python expressions
    @type EVAL_NAMESPACE: C{dict}
    """

    def __init__(self, **kwds):
        """
        Create an empty dataset.

        If keyword arguments are provided, they are interpreted as
        column definitions and are used to populate the dataset.

        @param kwds: I{column name} <-> I{vector} pairs
        @type  kwds: C{dict} mapping C{str}ings onto C{vector}s
        """

        self.__columns = {}
        self.__columnNames = []
        for name, vector in kwds.items():
            self.add(name, vector)

    def copy(self, deep=1):
        """
        Create a copy of this dataset.

        @param deep: make a deep copy of this dataset by making copies of all
                     of the vectors
        @type  deep: C{bool} (defaults to C{true})
        @return: copy of this dataset
        @rtype:  C{xdp.Data}
        """

        new = Data()

        if deep:
            for columnName in self.__columns:
                new.set(columName, self.__columns[name].copy())
        else:
            for columnName in self.__columns:
                new.set(columName, self.__columns[name])
        return new

    def columns(self):
        """
        Retrieve a list of the column names.

        @return: names of this dataset's columns
        @rtype:  C{list} of C{str}ings

        @depreciated: use C{getColumnNames} instead
        """

        return list(self.__columnNames)

    def getColumnNames(self):
        """
        Retrieve a list of the column names.

        @return: names of this dataset's columns
        @rtype:  C{list} of C{str}ings
        """

        return list(self.__columnNames)

    def has(self, name):
        """
        Check for the presence of a column.

        @param name: column to check for
        @type  name: C{str}ing
        @return: whether or not this dataset has a column C{name}
        @rtype:  C{bool}

        @depreciated: use C{hasColumn} instead
        """

        return self.__columns.has_key(name) or self.__getIndex(name) is not None

    def hasColumn(self, name):
        """
        Check for the presence of a column.

        @param name: column to check for
        @type  name: C{str}ing
        @return: whether or not this dataset has a column C{name}
        @rtype:  C{bool}
        """

        return self.__columns.has_key(name) or self.__getIndex(name) is not None

    def get(self, name):
        """
        Retrieve a column's data.

        C{name} may be the column's name, a column index, or a string
        containing a column index.

        @param name: column name
        @type  name: C{str}ing or C{int}
        @return: column data
        @rtype:  C{Numeric} vector

        @raise ColumnNameError: there is no column C{name} and C{name} is not
                                a valid column number

        @note: Column indices start at 1.

        @depreciated: use C{getColumn} instead
        """

        if self.__columns.has_key(name):
            return self.__columns[name]
        else:
            return self.__getByIndex(name)

    def getColumn(self, name):
        """
        Retrieve a column's data.

        C{name} may be the column's name, a column index, or a string
        containing a column index.

        @param name: column name
        @type  name: C{str}ing or C{int}
        @return: column data
        @rtype:  C{Numeric} vector

        @raise ColumnNameError: there is no column C{name} and C{name} is not
                                a valid column number

        @note: Column indices start at 1.
        """

        if self.__columns.has_key(name):
            return self.__columns[name]
        else:
            return self.__getByIndex(name)

    def add(self, name, vector):
        """
        Create a new column of data.

        @param name: new column's name
        @type  name: C{str}ing
        @param vector: new column's data
        @type  vector: C{Numeric} vector

        @raise ColumnNameError: the column C{name} already exists

        @depreciated: use C{addColumn} instead
        """

        if not self.__columns.has_key(name) and self.__getIndex(name) == None:
            self.__columns[name] = vector
            self.__columnNames.append(name)
        else:
            raise ColumnNameError(name)

    def addColumn(self, name, vector):
        """
        Create a new column of data.

        @param name: new column's name
        @type  name: C{str}ing
        @param vector: new column's data
        @type  vector: C{Numeric} vector

        @raise ColumnNameError: the column C{name} already exists

        @todo: check that the length of C{vector} matches the other columns
        @todo: check that C{name} is a valid column name
        """

        if not self.__columns.has_key(name) and self.__getIndex(name) == None:
            self.__columns[name] = vector
            self.__columnNames.append(name)
        else:
            raise ColumnNameError(name)

    def set(self, name, vector):
        """
        Assign data to a column.  If the column does not exist, it will be
        created.

        @param name: column name
        @type  name: C{str}ing
        @param vector: new column data
        @type  vector: C{Numeric} vector

        @depreciated: use C{setColumn} instead
        """

        if self.__columns.has_key(name):
            self.__columns[name] = vector
        elif self.__getIndex(name):
            self.__setByIndex(name, vector)
        else:
            self.__columns[name] = vector
            self.__columnNames.append(name)

    def setColumn(self, name, vector):
        """
        Assign data to a column.

        If the column does not exist, it will be created.

        @param name: column name
        @type  name: C{int} or C{str}ing
        @param vector: new column data
        @type  vector: C{Numeric} vector

        @todo: same checks as for C{addColumn}
        """

        if self.__columns.has_key(name):
            self.__columns[name] = vector
        elif self.__getIndex(name):
            self.__setByIndex(name, vector)
        else:
            self.__columns[name] = vector
            self.__columnNames.append(name)

    def remove(self, name):
        """
        Remove a column.

        @param name: column name to remove
        @type  name: C{str}ing

        @raise ColumnNameError: the column C{name} does not exist

        @depreciated: use C{removeColumn} instead
        """

        if self.__columns.has_key(name):
            del self.__columns[name]
            self.__columnNames.remove(name)
        else:
            self.__removeByIndex(name)

    def removeColumn(self, name):
        """
        Remove a column.

        @param name: column name or index to remove
        @type  name: C{int} or C{str}ing

        @raise ColumnNameError: the column C{name} does not exist
        """

        if self.__columns.has_key(name):
            del self.__columns[name]
            self.__columnNames.remove(name)
        else:
            self.__removeByIndex(name)

    def rename(self, name, newName):
        """
        Rename a column.

        @param name: column to rename
        @type  name: C{str}ing
        @param newName: new column name
        @type  newName: C{str}ing

        @raise TypeError: either C{name} or C{newName} is not a C{str}ing
        @raise ColumnNameError: the column C{name} does not exist

        @depreciated: use C{renameColumn} instead
        """

        if (not isinstance(name, types.StringType)
        or  not isinstance(newName, types.StringType)):
            raise TypeError('xdp.Data.rename() requires string arguments')

        if name != newName:
            self.set(newName, self.get(name))
            self.remove(name)

    def renameColumn(self, name, newName):
        """
        Rename a column.

        @param name: column to rename
        @type  name: C{str}ing
        @param newName: new column name
        @type  newName: C{str}ing

        @raise TypeError: either C{name} or C{newName} is not a C{str}ing
        @raise ColumnNameError: the column C{name} does not exist

        @todo: what is C{newName} already exists?
        """

        if (not isinstance(name, types.StringType)
        or  not isinstance(newName, types.StringType)):
            raise TypeError('xdp.Data.renameColumn() requires string arguments')

        if name != newName:
            self.set(newName, self.get(name))
            self.remove(name)

    def getMatrix(self):
        """
        Construct a matrix out of this dataset.

        @return: matrix containing all of this dataset's columns
        @rtype:  C{Numeric} matrix

        @note: Until C{addColumn} and C{setColumn} check that the size of
        column vectors is valid, this method could concievably throw a
        C{ValueError}.
        """

        columns = len(self.__columnNames)

        if not columns:
            return None

        rows = len(self.get(1))
        matrix = Numeric.zeros((rows, columns), Numeric.Float)

        for i in range(0, columns):
            matrix[:, i] = self.__columns[self.__columnNames[i]]
        return matrix

    def evaluate(self, expr):
        """
        Evaluate a Python expression using this dataset's column names

        Column names are prefixes by an at sign, and may contain the name
        (C{@If/@Io}) or number (C{@4/@2}).

        The following constants and functions are available for use within the
        expression:
         - C{int} S{->} C{int}
         - C{float} S{->} C{float}
         - C{complex} S{->} C{complex}
         - C{pi} S{->} C{math.pi}
         - C{e} S{->} C{math.e}
         - C{abs} S{->} C{Numeric.absolute}
         - C{arccos} S{->} C{Numeric.arccos}
         - C{arccosh} S{->} C{Numeric.arccosh}
         - C{arcsin} S{->} C{Numeric.arcsin}
         - C{arcsinh} S{->} C{Numeric.arcsinh}
         - C{arctan} S{->} C{Numeric.arctan}
         - C{arctanh} S{->} C{Numeric.arctanh}
         - C{cos} S{->} C{Numeric.cos}
         - C{cosh} S{->} C{Numeric.cosh}
         - C{exp} S{->} C{Numeric.exp}
         - C{log} S{->} C{Numeric.log}
         - C{log10} S{->} C{Numeric.log10}
         - C{power} S{->} C{Numeric.power}
         - C{sin} S{->} C{Numeric.sin}
         - C{sinh} S{->} C{Numeric.sinh}
         - C{sqrt} S{->} C{Numeric.sqrt}
         - C{tan} S{->} C{Numeric.tan}
         - C{tanh} S{->} C{Numeric.tanh}
        All of the functions from C{Numeric} are C{ufunc}s; they will operate
        correctly on vectors.

        @param expr: python expression
        @type  expr: C{str}ing
        @return: result of the expression
        @rtype:  I{any}

        @raise EvaluationError: the expression could not be compiled or caused
                                an exception during evaluation
        @raise ColumnNameError: the expression contained an invalid column
                                name or number
        """

        columnCount = len(self.__columnNames)
        namespace = self.EVAL_NAMESPACE.copy()
        for i in range(0, columnCount):
            name = self.__columnNames[i]
            vector = self.__columns[name]
            namespace['__C_%d' % (i+1)] = vector
            namespace['__C_%s' % name] = vector

        for column, _, name in self.COLNAMES.findall(expr):
            if not self.__columns.has_key(name):
                raise ColumnNameError(name)

        for column, _, number in self.COLNUMS.findall(expr):
            number = int(number)
            if number < 1 or number > columnCount:
                raise ColumnNameError(column)

        expr = self.COLNAMES.sub('__C_\g<name>', expr)
        expr = self.COLNUMS.sub('__C_\g<number>', expr)

        try:
            code = compile(expr, '<expr>', 'eval')
        except SyntaxError:
            raise EvaluationError('synatx error')

        for var in code.co_names:
            if not namespace.has_key(var):
                raise EvaluationError('unbound variable %r' % var)

        try:
            return eval(code, namespace)
        except:
            type, value, traceback = sys.exc_info()
            raise EvaluationError('%s <%s>' % (value, type))

    def __setattr__(self, name, value):
        """
        Allow columns to be set as attributes.

        @param name: column name (attribute name)
        @type  name: C{str}ing
        @param value: column data (attribute value)
        @type  value: C{Numeric} vector (I{any})

        @note: Columns must already exist before the can be updated using this
        method.

        @todo: same checks as for C{setColumn}
        """

        if name[0] != '_' and self.__columns.has_key(name):
            self.__columns[name] = value
        else:
            self.__dict__[name] = value

    def __getattr__(self, name):
        """
        Allow columns to be accessed as attributes.

        @param name: column name (attribute name)
        @type  name: C{str}ing

        @note: Columns must already exist before the can be fetched using this
        method.
        """

        if name[0] != '_' and self.__columns.has_key(name):
            return self.__columns[name]
        else:
            try:
                return self.__dict__[name]
            except KeyError:
                raise AttributeError(name)

    def __getIndex(self, name):
        """
        Convert a value from a column number into a column name.

        This method returns C{None} if C{name} is not a valid column number.

        @param name: column number
        @type  name: C{int} or C{str}ing
        @return: column name
        @rtype:  C{str}ing or C{None}

        Attempt to convert the object 'name' into a valid column name by
        interpreting it as a column number.  A return value of None indicates
        that 'name' was not a valid column number.
        """

        try:
            index = int(name) - 1
        except ValueError:
            return None
        else:
            if index < 0 or index > len(self.__columnNames):
                return None
            else:
                return self.__columnNames[index]

    def __getByIndex(self, index):
        """
        Retrieve data by column number.

        @param index: column number
        @type  index: C{int} or C{str}ing
        @return: column data
        @rtype:  C{Numeric} vector

        @raise ColumnNameError: the column number supplied was invalid
        """

        name = self.__getIndex(index)

        if name is None:
            raise ColumnNameError(index)
        else:
            return self.__columns[name]

    def __setByIndex(self, index, vector):
        """
        Assign data by column number.

        @param index: column number
        @type  index: C{int} or C{str}ing
        @param vector: new column data
        @type  vector: C{Numeric} vector

        @raise ColumnNameError: the column number supplied was invalid
        """

        name = self.__getIndex(index)

        if name is None:
            raise ColumnNameError(index)
        else:
            self.__columns[name] = vector

    def __removeByIndex(self, index):
        """
        Remove a column by its column number.

        @param index: column number
        @type  index: C{int} or C{str}ing

        @raise ColumnNameError: the column number supplied was invalid
        """
        name = self.__getIndex(index)

        if name is None:
            raise ColumnNameError(index)
        else:
            del self.__columns[name]

    COLNAMES = re.compile(
                r'(?P<column>(@|\$)(?P<name>[a-zA-Z_][a-zA-Z0-9_]*))')

    COLNUMS = re.compile(
                r'(?P<column>(@|\$)(?P<number>\d+))')

    EVAL_NAMESPACE = {
        'int': int,
        'float': float,
        'complex': complex,
        'pi': math.pi,
        'e': math.e,
        'abs': Numeric.absolute,
        'arccos': Numeric.arccos,
        'arccosh': Numeric.arccosh,
        'arcsin': Numeric.arcsin,
        'arcsinh': Numeric.arcsinh,
        'arctan': Numeric.arctan,
        'arctanh': Numeric.arctanh,
        'cos': Numeric.cos,
        'cosh': Numeric.cosh,
        'exp': Numeric.exp,
        'log': Numeric.log,
        'log10': Numeric.log10,
        'power': Numeric.power,
        'sin': Numeric.sin,
        'sinh': Numeric.sinh,
        'sqrt': Numeric.sqrt,
        'tan': Numeric.tan,
        'tanh': Numeric.tanh
    }


__all__ = ['Data']
