# Name: xdp/io/xafs.py
# Purpose: I/O support for MR-CAT XAFS data files
# Author: Ken McIvor <mcivor@iit.edu>
#
# Copyright 2003-2005 Illinois Institute of Technology
#
# See the file "LICENSE" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.

"""
Read and write files using the MR-CAT XAFS file format, version 0.3.
"""


import Numeric

import xdp
import _matrixio


class Xafs:
    """
    The contents of the header of an MR-CAT XAFS file.

    Default field values:
     - FileInfo:     C{''}
     - RingEnergy:   C{[0.0, 'eV']}
     - E0:           C{0.0}
     - NumRegions:   C{0}
     - SRB:          C{[]}
     - SRSS:         C{[]}
     - SPP:          C{[]}
     - SettlingTime: C{0.0}
     - Offsets:      C{[]}
     - Gains:        C{[]}
     - Comment1:     C{''}
     - Comment2:     C{''}
     - Comment3:     C{''}

    @ivar FileInfo: description of the dataset
    @type FileInfo: C{str}ing
    @ivar RingEnergy: header field
    @type RingEnergy: C{2-tuple} of the form C{energy, units}
    @ivar E0: edge energy
    @type E0: C{float}
    @ivar NumRegions: header field
    @type NumRegions: C{int}
    @ivar SRB: header field
    @type SRB: C{list} of C{int}s
    @ivar SRSS: header field
    @type SRSS: C{list} of C{int}s
    @ivar SPP: header field
    @type SPP: C{list} of C{int}s
    @ivar SettlingTime: header field
    @type SettlingTime: C{float}
    @ivar Offsets: header field
    @type Offsets: C{list} of C{floats}
    @ivar Gains: header field
    @type Gains: C{list} of C{floats}
    @ivar Comment1: comment line 1
    @type Comment1: C{str}ing
    @ivar Comment2: comment line 2
    @type Comment2: C{str}ing
    @ivar Comment3: comment line 3
    @type Comment3: C{str}ing

    @todo: better descriptions than I{header field}

    @todo: accessors that check the field values for validity
    """

    def __init__(self):
        """Create an empty XAFS header.
        """
        self.FileInfo = ''
        self.RingEnergy = [0.0, 'eV']
        self.E0 = 0.0
        self.NumRegions = 0
        self.SRB = []
        self.SRSS = []
        self.SPP = []
        self.SettlingTime = 0.0
        self.Offsets = []
        self.Gains = []
        self.Comment1 = ''
        self.Comment2 = ''
        self.Comment3 = ''

    def write(self, data, stream):
        """
        Write a dataset to an output stream in the MR-CAT XAFS file format,
        using this header's fields.

        @param data: dataset
        @type  data: C{Data}
        @param stream: output stream
        @type  stream: C{file}

        @note: The output is written in the MRCAT XAFS format.

        @todo: check values thoroughly while outputting them
        """
        _xafs_writeHeader(self, stream)
        stream.write(' '.join(data.columns()))
        stream.write('\n')
        _matrixio.write(stream, data.getMatrix())


def readXafsFile(stream):
    """
    Read an XAFS header and data from an input stream.

    @param stream: input stream
    @type  stream: C{file}
    @return: header and data
    @rtype:  C{2-tuple} of the form C{(header, data)}

    @raise IOError: the XAFS header or data was malformed.
    """
    header = Xafs()
    columnNames = _xafs_readHeader(stream, header)
    data = xdp.Data()

    try:
        matrix = _matrixio.read(stream)
    except _matrixio.error, e:
        raise IOError(str(e))

    nCols = matrix.shape[1]
    nNames = len(columnNames)

    if nNames < nCols:
        N = nNames
        extraCols = nCols - nNames
    else:
        N = nCols
        extraCols = 0

    for i in range(0, N):
        data.add(columnNames[i], matrix[:, i])

    for i in range(N, extraCols + N):
        data.add('column%d' % i, matrix[:, i])

    return header, data

#
# Implementation functions
#


_xafs_RING = 1
_xafs_FLOAT = 2
_xafs_ARRAY = 3


_xafs_fields = [
    ('RingEnergy', 'Ring energy', _xafs_RING),
    ('E0', 'E0', _xafs_FLOAT),
    ('NumRegions', 'NUM_REGIONS', _xafs_FLOAT),
    ('SRB', 'SRB', _xafs_ARRAY),
    ('SRSS', 'SRSS', _xafs_ARRAY),
    ('SPP', 'SPP', _xafs_ARRAY),
    ('SettlingTime', 'Settling time', _xafs_FLOAT)
]


def _xafs_readHeader(stream, header):
    line = stream.readline()

    if line[0:10] != 'MRCAT_XAFS':
        raise IOError('xafs header line missing')

    pieces = line.split()
    if len(pieces) != 3:
        raise IOError('malformed header')
    else:
        version, fileType = pieces[1:3]

    if version != 'V0.3':
        raise IOError('unsupported version of the XAFS format')

    if fileType != 'Datafile':
        raise IOError('unsupported XAFS file type "%s"' % type)

    header.FileInfo = stream.readline().strip()

    for attributeName, fieldName, fieldType in _xafs_fields:
        setattr(header, attributeName,
            _xafs_parseHeaderField(stream.readline(), fieldName, fieldType))

    if stream.readline().strip() or stream.readline().strip():
        raise IOError('malformed header')

    for fieldName in ('Offsets', 'Gains'):
        setattr(header, fieldName,
            _xafs_parseHeaderField(stream.readline(), fieldName, _xafs_ARRAY))

    header.Comment1 = stream.readline().strip()
    header.Comment2 = stream.readline().strip()
    header.Comment3 = stream.readline().strip()

    if stream.readline().strip() != ('-' * 96):
        raise IOError('header separator missing')

    columns = stream.readline().split()

    return columns


def _xafs_parseHeaderField(line, fieldName, type):
    pieces = line.split('=', 1)
    if len(pieces) != 2:
        raise IOError('malformed field "%s"' % fieldName)
    else:
        name = pieces[0].strip()
        value = pieces[1].strip()

    if name != fieldName:
        raise IOError('unexpected field %s (wanted %s)' % (name, fieldName))

    if type == _xafs_RING:
        try:
            value, units = value.split()
            value = float(value)
        except ValueError:
            raise IOError('malformed field "%s"' % fieldName)

        value = [value, units]

    elif type == _xafs_FLOAT:
        try:
            value = float(value)
        except ValueError:
            raise IOError('malformed field "%s"' % fieldName)

    elif type == _xafs_ARRAY:
        value = value.split()
    return value


def _xafs_writeHeader(header, stream):
    stream.write('MRCAT_XAFS V0.3 Datafile\n')
    stream.write(header.FileInfo)
    stream.write('\n')

    for (attributeName, fieldName, fieldType) in _xafs_fields:
        _xafs_writeHeaderField(header, stream, attributeName, fieldName,
            fieldType)

    stream.write('\n\n')

    for fieldName in ('Offsets', 'Gains'):
        _xafs_writeHeaderField(header, stream, fieldName, fieldName,
            _xafs_ARRAY)

    stream.write(header.Comment1)
    stream.write('\n')
    stream.write(header.Comment2)
    stream.write('\n')
    stream.write(header.Comment3)
    stream.write('\n' + ('-' * 96) + '\n')


def _xafs_writeHeaderField(header, stream, attributeName, fieldName, type):
    if type == _xafs_RING:
        value, type = getattr(header, attributeName)
        line = '%s= %s %s\n' % (fieldName, value, type)
    elif type == _xafs_FLOAT:
        line = '%s= %s\n' % (fieldName, getattr(header, attributeName))
    elif type == _xafs_ARRAY:
        line = '%s= %s\n' % (fieldName,
            ' '.join([str(x) for x in getattr(header, attributeName)]))
    stream.write(line
    )
