/* Name: _bezier.c
 * Purpose: a fast C implementaton of bezier curves for Python
 * Author: Ken McIvor <mcivken@iit.edu>
 *
 * Copyright 2003-2004 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */


#include <math.h>
#include <stdlib.h>

#include <Python.h>
#include <Numeric/arrayobject.h>
#define _BEZIER_MODULE
#include "_bezier.h"


static char *__version__ = "1.1";
static char *__rcsid__ = "$Id$";
static char *__source__ = "$Source$";


#define pyRaise(type, mesg) { PyErr_SetString(type, mesg); return NULL; }


static PyObject *_bezier_pyCoeff(PyObject *self, PyObject *args)
{
    PyArrayObject *coeff;

    if (!PyArg_ParseTuple(args, "O!", &PyArray_Type, &coeff))
        return NULL;

    if (coeff->nd != 1)
        pyRaise(PyExc_ValueError, "array must be a vector");
    
    if (coeff->descr->type_num != PyArray_DOUBLE)
        pyRaise(PyExc_ValueError, "array must be of type Numeric.Float64");

    if (coeff->dimensions[0] < 2)
        pyRaise(PyExc_ValueError,
            "cannot generate coefficients for curves of orders less than two");

    Py_INCREF(coeff);

    _bezier_coeff(coeff);

    Py_DECREF(coeff);
    Py_INCREF(Py_None);
    return Py_None;
}


static PyObject *_bezier_pyBint(PyObject *self, PyObject *args)
{
    double u;
    PyArrayObject *xcont, *ycont, *coeff, *out;

    if (!PyArg_ParseTuple(args, "dO!O!O!O!", &u, &PyArray_Type, &xcont, \
        &PyArray_Type, &ycont, &PyArray_Type, &coeff, &PyArray_Type, &out))
        return NULL;

    if (xcont->nd != 1)
        pyRaise(PyExc_ValueError, "x argument must be a vector");

    if (ycont->nd != 1)
        pyRaise(PyExc_ValueError, "y argument must be a vector");

    if (coeff->nd != 1)
        pyRaise(PyExc_ValueError, "coeff argument must be a vector");

    if (out->nd != 1)
        pyRaise(PyExc_ValueError, "out argument must be a vector");

    if (xcont->descr->type_num != PyArray_DOUBLE)
        pyRaise(PyExc_ValueError, "x argument must be of type Numeric.Float64");

    if (ycont->descr->type_num != PyArray_DOUBLE)
        pyRaise(PyExc_ValueError, "y argument must be of type Numeric.Float64");

    if (coeff->descr->type_num != PyArray_DOUBLE)
        pyRaise(PyExc_ValueError, \
            "coeff argument must be of type Numeric.Float64");

    if (out->descr->type_num != PyArray_DOUBLE)
        pyRaise(PyExc_ValueError, \
            "out argument must be of type Numeric.Float64");

    if (xcont->dimensions[0] != ycont->dimensions[0])
        pyRaise(PyExc_ValueError, "x and y vectors must be the same length");

    if (xcont->dimensions[0] != coeff->dimensions[0])
        pyRaise(PyExc_ValueError,
            "the number of x/y pairs must equal curve order+1");

    if (out->dimensions[0] != 2)
        pyRaise(PyExc_ValueError, "out vector must have a shape of (2,)");

    Py_INCREF(xcont);
    Py_INCREF(ycont);
    Py_INCREF(coeff);
    Py_INCREF(out);

    _bezier_bint(u, xcont, ycont, coeff, out);

    Py_DECREF(xcont);
    Py_DECREF(ycont);
    Py_DECREF(coeff);
    Py_DECREF(out);

    Py_INCREF(Py_None);
    return Py_None;
}


static void _bezier_coeff(PyArrayObject *c)
{
    double *d;
    int n, k, i;

    n = c->dimensions[0] - 1;

    for (k = 0; k <= n; k++) {
        d = (double *)(c->data + k*c->strides[0]);
        *d = 1.0;
    
        for (i = n; i > k; i--)
            *d = *d * i;

        for (i = n-k; i > 1; i--)
            *d = *d / i;
    }
}


static void _bezier_bint(double u, PyArrayObject *xcont, PyArrayObject *ycont,
    PyArrayObject *coeff, PyArrayObject *out)
{
    int n, k;
    double blend, X, Y;

    X = Y = 0.0;
    n = xcont->dimensions[0] - 1;

    for (k = 0; k <= n; k++) {
        blend = *(double *)(coeff->data + k*coeff->strides[0]) * pow(u, k) \
            * pow(1.0-u, n-k);
        X += *(double *)(xcont->data + k*xcont->strides[0]) * blend;
        Y += *(double *)(ycont->data + k*ycont->strides[0]) * blend;
    }

    *(double *)(out->data) = X;
    *(double *)(out->data + out->strides[0]) = Y;
}


static PyMethodDef _bezierMethods[] = {
    {"coeff", _bezier_pyCoeff, METH_VARARGS,
        "coeff(Array) -- populates Array with coefficients"},
    {"bint", _bezier_pyBint, METH_VARARGS,
        "bint(u, x, y, coeff, out) -- out[:] = results of bezier interpolation"},
    {NULL, NULL, 0, NULL}
};


void init_bezier(void)
{
    PyObject *m, *s, *dict;
    static void *_bezier_API[_bezier_API_pointers];
    PyObject *c_api_object;

    m = Py_InitModule("_bezier", _bezierMethods);

    _bezier_API[_bezier_pyCoeff_NUM] = (void *) _bezier_pyCoeff;
    _bezier_API[_bezier_pyBint_NUM] = (void *) _bezier_pyBint;
    _bezier_API[_bezier_coeff_NUM] = (void *) _bezier_coeff;
    _bezier_API[_bezier_bint_NUM] = (void *) _bezier_bint;
    c_api_object = PyCObject_FromVoidPtr((void *) _bezier_API, NULL);

    if (c_api_object != NULL) {
        PyObject *d = PyModule_GetDict(m);

        PyDict_SetItemString(d, "_C_API", c_api_object);
        Py_DECREF(c_api_object);
    }

    dict = PyModule_GetDict(m);

    s = Py_BuildValue("s", __version__);
    if (s != NULL)
        PyDict_SetItemString(dict, "__version__", s);

    s = Py_BuildValue("s", __rcsid__);
    if (s != NULL)
        PyDict_SetItemString(dict, "__rcsid__", s);

    s = Py_BuildValue("s", __source__);
    if (s != NULL)
        PyDict_SetItemString(dict, "__source__", s);

    import_array();

    if (PyErr_Occurred())
        Py_FatalError("can't initialize module xdp.interpolate._bezier");
}
